#include <iostream>
#include <string>
#include <fstream>
#include <cstdlib>  // for NULL

#include <string.h>

#include "mpi.h"

#include "phRead.h"
#include "subsample.h"

using std::endl;
using std::cout;
using std::cin;
using std::ofstream;

enum arg_enum_t {arg_int_type=1, arg_double_type, arg_char_type};

int procArgs(int argc, char *argv[], const char *token, int array_size, arg_enum_t array_type, void *array_values);

int main(int argc, char *argv[])
{

  MPI_Init(&argc,&argv);

  int numprocs;
  MPI_Comm_size(MPI_COMM_WORLD, &numprocs);

  int myrank;
  MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

  int displayHelp = 0;
  for(int iArgc=1; iArgc<argc; iArgc++) {
    if(!strcmp(argv[iArgc],"-h")) {
      if(myrank==0) {
        cout << endl;
        cout << "  HELP requested (by using \"-h\" in arguments) -- " << endl;
      }
      displayHelp = 1;
    }
  }

  // inputs
  int numparts;
  int start_step_number; int incr_step_size; int end_step_number;
  int sub_option;

  int err_procArgs=0;
  if(!displayHelp) {
    err_procArgs += procArgs(argc,argv,"-np",1,arg_int_type,(void*)&numparts);

    int ts_array_size=3; arg_enum_t ts_array_type=arg_int_type; int ts_array_values[3];
    err_procArgs += procArgs(argc,argv,"-ts",ts_array_size,ts_array_type,(void*)ts_array_values);
    start_step_number=ts_array_values[0];incr_step_size=ts_array_values[1];end_step_number=ts_array_values[2];

    err_procArgs += procArgs(argc,argv,"-sub_option",1,arg_int_type,(void*)&sub_option);
  }

  if (myrank==0 && (displayHelp || err_procArgs)) {
    cout << endl;
    cout << "  usage : " << endl;
    cout << "     <executable>" << endl;
    cout << "     -np <parts>" << endl;
    cout << "     -ts <start_step_number> <incr_step_size> <end_step_number>" << endl;
    cout << "     -sub_option <subsample_option>" << endl;
    cout << endl;
    exit(1);
  }

  if(myrank==0)
    cout << "  sub-sample [option:" << sub_option << "] between time-steps [" << start_step_number << "," << end_step_number << "] at every " << incr_step_size << " step ..." << endl;

  int isize, iarray[20], asize; // used for phasta files (in phastaIO format)

  char fprefix[64];
  sprintf(fprefix,"%d-procs_case",numparts);

  std::string integer_str = "integer";
  std::string double_str = "double";

  int numNodes, numSD, numSizeILWork, numSolVars;

  if(numparts<numprocs) {
    if(myrank==0)
      cout<<"\n Error : numparts less than numprocs..."<<endl;
    exit(1);
  }

  // check range of parts that needs to be processed by this rank
  int numparts_l = numparts/numprocs;

  // note myrank starts from 0 (lowest one is 0)
  // startpart will begin from 1 on lowest rank
  // (also order of operations is imp. as numparts_l will be changed on last rank/proc. as below)
  int startpart = myrank*numparts_l+1;

  // last rank/proc. may get extra partial number of parts
  if(myrank+1==numprocs)
    numparts_l = numparts-(numprocs-1)*numparts_l;

  // startpart begins from 1
  // (order of operations is imp. as numparts_l is corrected on last rank/proc.)
  int endpart_plus1 = startpart+numparts_l;

  // special reduced case
  if(numprocs==1 && numparts==1) {
    startpart=0; endpart_plus1=1; // not considering single-proc. case
  }

  cout<<"  ["<<myrank<<"] parts range : "<<startpart<<" - "<<endpart_plus1-1<<endl;

  for(int iPart=startpart;iPart<endpart_plus1;iPart++){
    char GeomBCFileName[256];
    sprintf(GeomBCFileName,"%s/geombc.dat.%d",fprefix,iPart);

    // read coordinates
    char coordFieldTag[64];
    sprintf(coordFieldTag,"co-ordinates");
    isize=2;
    readParametersFromFile(GeomBCFileName,coordFieldTag,double_str.c_str(),isize,iarray);
    numNodes=iarray[0]; numSD=iarray[1];
    double *f_coord = new double[numNodes*numSD];
    asize=numNodes*numSD;
    readArrayFromFile(GeomBCFileName,coordFieldTag,double_str.c_str(),isize,iarray,asize,(void *)f_coord);

    // F to C
    double *c_coord = new double[numNodes*numSD];
    for(int i = 0; i< numNodes; i++){
      for( int j=0; j< numSD; j++){
        c_coord[i*numSD+j] = f_coord[j*numNodes+i];
      }
    }
    delete [] f_coord;

    int *flag_for_points = new int[numNodes];
    for(int i = 0; i< numNodes; i++)
      flag_for_points[i] = 0;

    // index of dofs (subsample):
    //   whose values are extracted from temporal snapshots
    subsample_idx(sub_option,numNodes,numSD,c_coord,flag_for_points);

    // read ilwork
    char ilworkFieldTag[64];
    sprintf(ilworkFieldTag,"ilwork");
    isize=1;
    readParametersFromFile(GeomBCFileName,ilworkFieldTag,integer_str.c_str(),isize,iarray);
    numSizeILWork=iarray[0];
    int *ilwork = new int[numSizeILWork];
    asize=numSizeILWork;
    readArrayFromFile(GeomBCFileName,ilworkFieldTag,integer_str.c_str(),isize,iarray,asize,(void *)ilwork);

    // process ilwork
    {
      int il_numents = ilwork[0];
      int il_count=0;
      for(int i=0; i<il_numents; i++) {
        int il_type = ilwork[il_count+2];
        int il_numseg = ilwork[il_count+4];
        int il_count2 = il_count;
        il_count += 4+2*il_numseg;
        if(il_type==1) // master/owner copy
          continue;

        // il_type==0 : non-owner copy then turn OFF entry in 'flag_for_points'
        for(int j=0; j<il_numseg; j++) {
          int il_idx = il_count2+3+2*(j+1); // for j=0 it is 'il_count2+5'
          int il_beginseg = ilwork[il_idx];
          int il_lenseg = ilwork[il_idx+1];
          for(int k=0; k<il_lenseg; k++) {
            int il_dof_idx = il_beginseg+k-1; // F to C index (C starts from 0)
            flag_for_points[il_dof_idx] = 0;
          }
        }
      }
    }

    delete [] ilwork;

    for(int iStep=start_step_number; iStep<=end_step_number; iStep=iStep+incr_step_size) {
      char SolutionFileName[256];
      sprintf(SolutionFileName,"%s/restart.%d.%d",fprefix,iStep,iPart);

      //read solution
      char solFieldTag[64];
      sprintf(solFieldTag,"solution");
      isize=3;
      readParametersFromFile(SolutionFileName,solFieldTag,double_str.c_str(),isize,iarray);
      numNodes=iarray[0]; numSolVars=iarray[1];
      double *f_solution = new double[numNodes*numSolVars];
      asize=numNodes*numSolVars;
      readArrayFromFile(SolutionFileName,solFieldTag,double_str.c_str(),isize,iarray,asize,(void *)f_solution);

      // F to C
      double *c_solution = new double[numNodes*numSolVars];
      for(int i = 0; i< numNodes; i++){
        for( int j=0; j< numSolVars; j++){
          c_solution[i*numSolVars+j] = f_solution[j*numNodes+i];
        }
      }
      delete [] f_solution;

      char subSampleSolutionFileName[256];
      sprintf(subSampleSolutionFileName,"subsample.%d.%d",iStep,iPart);
      ofstream subsample_outfile(subSampleSolutionFileName);
      if(iPart==1)
        cout<<"  writing subsampled solution in file: "<<subSampleSolutionFileName<<" and so on ..."<<endl;
      for(int i=0; i<numNodes; i++) {
        if(!flag_for_points[i])
          continue;

        subsample_outfile << i;
        // TODO: do we need coords? (probably not at each time-step)
        for(int j=0; j<numSolVars; j++)
          subsample_outfile << " " << c_solution[i*numSolVars+j];
        subsample_outfile << endl;
      }
      subsample_outfile.close();

      delete [] c_solution;
    } // iStep-loop

    delete [] c_coord;
    delete [] flag_for_points;
  } // iPart-loop

  MPI_Barrier(MPI_COMM_WORLD);

  MPI_Finalize();

  return 0;
}

int procArgs(int argc, char *argv[], const char *token, int array_size, arg_enum_t array_type, void *array_values) {

  for(int iArgc=0; iArgc<argc; iArgc++) {
    if(!strcmp(argv[iArgc],token)) {
      if(iArgc+array_size>=argc) // not enough arguments
        break;

      if(array_type==arg_int_type) { // integer type
        for(int iArgc2=0; iArgc2<array_size; iArgc2++) {
          int index = iArgc+iArgc2+1;
          ((int*)array_values)[iArgc2] = atoi(argv[index]);
        }
        return 0;
      }
      else if(array_type==arg_double_type) { // double type
        for(int iArgc2=0; iArgc2<array_size; iArgc2++) {
          int index = iArgc+iArgc2+1;
          ((double*)array_values)[iArgc2] = atof(argv[index]);
        }
        return 0;
      }
      else if(array_type==arg_char_type) { // char type
        for(int iArgc2=0; iArgc2<array_size; iArgc2++) {
          int index = iArgc+iArgc2+1;
          ((std::string*)array_values)[iArgc2] = argv[index];
        }
        return 0;
      }
      else // error
        break;

    }
  }

  return 1; // error
}
